# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Pyqrlew'
copyright = '2023, Sarus Technologies'
author = 'Sarus Technologies'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    # 'myst_parser',
    'myst_nb',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'sphinxcontrib.bibtex',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# Enable numref
numfig = True
bibtex_bibfiles = ['qrlew.bib']

# MyST conf
myst_enable_extensions = [
    'dollarmath',
    'amsmath',
]
myst_dmath_allow_labels = True
myst_dmath_double_inline = True

# Controls notebook execution
nb_execution_mode = 'cache'#'off'/'cache'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'

html_static_path = ['_static']

html_theme_options = {
    'navbar_start': ['navbar-logo'],
    'navbar_center': ['navbar-nav'],
    'navbar_end': ['navbar-icon-links'],
    'navbar_persistent': ['search-button'],
    'icon_links': [
        {
            'name': 'GitHub',
            'url': 'https://github.com/Qrlew',
            'icon': 'fa-brands fa-github',
            'type': 'fontawesome',
        },
        {
            'name': 'Web Page',
            'url': 'https://qrlew.github.io/',
            'icon': 'fa-solid fa-house',
            'type': 'fontawesome',
        },
        {
            'name': 'Discord',
            'url': 'https://discord.gg/JbvSPgyp',
            'icon': 'fa-brands fa-discord',
            'type': 'fontawesome',
        },
        {
            'name': 'X',
            'url': 'https://x.com/sarus_tech',
            'icon': 'fa-brands fa-twitter',
            # The default for `type` is `fontawesome` so it is not actually required in any of the above examples as it is shown here
        },
    ],
}


from pathlib import Path
from typing import List
import pytest


@pytest.fixture
def tests_path() -> Path:
    return Path(__file__).parent

@pytest.fixture
def queries_path(tests_path) -> Path:
    """Provide the path to test queries"""
    return tests_path / 'queries' / 'base_queries.sql'

@pytest.fixture
def queries(queries_path) -> List[str]:
    with open(queries_path, 'r') as f:
        return [query for query in f if not query.startswith('--')]



# Import rust code
from pyqrlew.pyqrlew import *

__doc__ = pyqrlew.__doc__
if hasattr(pyqrlew, "__all__"):
    __all__ = pyqrlew.__all__

# Import python modules
from pyqrlew import io
from pyqrlew import tester


import logging
import pyqrlew as qrl
from pyqrlew.io import PostgreSQL
from pyqrlew.io.dataset import dataset_from_database

logging.basicConfig(level=logging.DEBUG)

db = PostgreSQL()
ds = db.extract()


from typing import List
import numpy as np
import pandas as pd
from ..io.postgresql import EmptyPostgreSQL
from pyqrlew.io.dataset import dataset_from_database
from sqlalchemy import text
from enum import Enum

class Distribution(Enum):
    LAPLACE = "laplace"
    UNIFORM = "uniform"
    GAUSSIAN = "gaussian"
    HALTON = "halton"

class ColumnSpec:
    def __init__(self, name: str, distribution: Distribution, nan_probability: float = 0.0, **kwargs):
        self.name = name
        self.distribution = distribution
        self.nan_probability = nan_probability
        self.kwargs = kwargs

class StochasticDatabase:
    def __init__(self, schema_name, dbname, user, password, port):
        db = EmptyPostgreSQL(dbname, user, password, port)
        self.schema_name = schema_name
        self.engine = db.engine()
        with self.engine.connect() as connection:
            connection.execute(f"CREATE SCHEMA IF NOT EXISTS {schema_name}")

    def create_table(self, name: str, size: int, column_specs: List[ColumnSpec], index_is_unique=True):
        data = RandomTableGenerator(size).create_table(column_specs)
        if index_is_unique:
            data["id"] = np.arange(size)
        else:
            data["id"] = [
                int(np.random.normal(loc = size / 2, scale = size / 4))
                for _ in range(size)
            ]
        data.to_sql(name, schema=self.schema_name, con=self.engine, if_exists='replace', index=False)

    def dataset(self):
        return dataset_from_database(self.schema_name, self.engine, self.schema_name)

    def eval(self, relation) -> list:
        return self.execute(relation.to_query())

    def execute(self, query: str) -> list:
        with self.engine.connect() as conn:
            result = conn.execute(text(query)).all()
        return result

class RandomTableGenerator:
    def __init__(self, size: int):
        self.size = size

    def create_table(self, column_specs: List[ColumnSpec]) -> pd.DataFrame:
        table_data = {
            spec.name: self._generate_numbers(spec)
            for spec in column_specs
        }
        return pd.DataFrame(table_data)

    def _generate_numbers(self, spec: ColumnSpec) -> np.ndarray:
        distribution = spec.distribution
        kwargs = spec.kwargs
        nan_probability = spec.nan_probability
        print

        random_numbers = np.zeros(self.size)
        if distribution == Distribution.LAPLACE:
            random_numbers =  np.random.laplace(**kwargs, size=self.size)
        elif distribution == Distribution.UNIFORM:
            random_numbers =  np.random.uniform(**kwargs, size=self.size)
        elif distribution == Distribution.GAUSSIAN:
            random_numbers =  np.random.normal(**kwargs, size=self.size)
        elif distribution == Distribution.HALTON:
            random_numbers =  self.halton_sequence(**kwargs)
        if  nan_probability > 0:
            random_numbers =  self.add_nan(random_numbers, nan_probability)
        return random_numbers

    def add_nan(self, random_numbers, nan_probability=0.1):
        nan_mask = random_numbers < nan_probability
        random_numbers[nan_mask] = np.nan
        return random_numbers

    def halton_sequence(self, base=2, low=0, high=1):
        sequence = np.zeros(self.size)

        for i in range(self.size):
            f = 1.0
            r = 0.0
            index = i + 1

            while index > 0:
                f /= base
                r += f * (index % base)
                index = index // base

            sequence[i] = r

        return low + (high - low) * sequence



import pyqrlew as pq
import pyqrlew.tester as tester

DBNAME: str = 'pyqrlew-db'
USER: str = 'postgres'
PASSWORD: str = 'pyqrlew-db'
PORT: str = 5433

def test_sum(database, stochastic_tester, n_sim, distributions, name, row_privacy=True):
    tablename = 'table1' if row_privacy else 'table2'
    rp = 'row_privacy' if row_privacy else 'multi_rows'

    def write_inner_query(d, a, b):
        return f"SELECT {d} as x FROM testing.{tablename} WHERE {d} > {a} AND {d} < {b}"

    for (d, a, b) in distributions:
        inner_query = write_inner_query(d, a, b)
        query = f"SELECT SUM(x) AS my_sum FROM ({inner_query}) AS my_table"
        adj_query = lambda id: f"SELECT SUM(x) AS my_sum FROM ({inner_query} AND id != {id}) AS my_table"
        tester.run_test(database, stochastic_tester, query=query, adj_query_func=adj_query, name=f'{name}_{d}_{rp}', n_sim=n_sim)

def test_avg(database, stochastic_tester, n_sim, distributions, name, row_privacy=True):
    tablename = 'table1' if row_privacy else 'table2'
    rp = 'row_privacy' if row_privacy else 'multi_rows'

    def write_inner_query(d, a, b):
        return f"SELECT {d} as x FROM testing.{tablename} WHERE {d} > {a} AND {d} < {b}"

    for (d, a, b) in distributions:
        inner_query = write_inner_query(d, a, b)
        query = f"SELECT AVG(x) AS my_sum FROM ({inner_query}) AS my_table"
        adj_query = lambda id: f"SELECT AVG(x) AS my_sum FROM ({inner_query} AND id != {id}) AS my_table"
        tester.run_test(database, stochastic_tester, query=query, adj_query_func=adj_query, name=f'{name}_{d}_{rp}', n_sim=n_sim)

def test_count(database, stochastic_tester, n_sim, name, row_privacy=True):
    tablename = 'table1' if row_privacy else 'table2'
    rp = 'row_privacy' if row_privacy else 'multi_rows'

    query = f"SELECT COUNT(x) AS my_sum FROM (SELECT laplace AS x FROM testing.{tablename}) AS my_table"
    adj_query = lambda id: f"SELECT COUNT(x) AS my_sum FROM (SELECT laplace AS x FROM testing.{tablename} WHERE id != {id}) AS my_table"
    tester.run_test(database, stochastic_tester, query=query, adj_query_func=adj_query, name=f'{name}_{rp}', n_sim=n_sim)

def test_joins(database, stochastic_tester, n_sim, name):
    inner_query = "SELECT t1.halton + 2 * t2.halton as x FROM testing.table1 AS t1 JOIN testing.table2 AS t2 ON t1.id = t2.id WHERE t1.halton > -1 AND t1.halton < 1 AND t2.halton > -1 AND t2.halton < 1"
    query = f"SELECT SUM(x) AS my_sum FROM ({inner_query}) AS my_table"
    adj_query = lambda id: f"SELECT SUM(x) AS my_sum FROM ({inner_query} AND id != {id}) AS my_table"
    tester.run_test(database, stochastic_tester, query=query, adj_query_func=adj_query, name=name, n_sim=n_sim)

if __name__ == "__main__":
    database = tester.StochasticDatabase('testing', DBNAME, USER, PASSWORD, PORT)
    print('creating tables ...')
    size = 1000
    ## create table
    column_specs = [
        tester.ColumnSpec(name="laplace", distribution=tester.Distribution.LAPLACE, loc=0, scale=1),
        tester.ColumnSpec(name="uniform", distribution=tester.Distribution.UNIFORM, low=-1, high=1),
        tester.ColumnSpec(name="gaussian", distribution=tester.Distribution.GAUSSIAN, loc=0, scale=1),
        tester.ColumnSpec(name="halton", distribution=tester.Distribution.HALTON, base=2, low=-1, high=1),
    ]
    database.create_table("table1", size, column_specs, True)
    column_specs = [
        tester.ColumnSpec(name="laplace", distribution=tester.Distribution.LAPLACE, loc=0, scale=1),
        tester.ColumnSpec(name="uniform", distribution=tester.Distribution.UNIFORM, low=-1, high=1),
        tester.ColumnSpec(name="gaussian", distribution=tester.Distribution.GAUSSIAN, loc=0, scale=1),
        tester.ColumnSpec(name="halton", distribution=tester.Distribution.HALTON, base=2, low=-1, high=1),
    ]
    database.create_table("table2", size, column_specs, False)
    print('done')
    privacy_unit = [
        ("table1", [], "id"),
        ("table2", [], "id"),
    ]
    synthetic_data = [
        (["testing", "table1"], ["testing", "table1"]),
        (["testing", "table2"], ["testing", "table2"]),
    ]

    distributions = [
        # ('laplace', -2.5, 2.5),
        # ('gaussian', -2., 2.),
        # ('uniform', -1., 1.),
        ('halton', -1., 1.),
    ]
    row_privacy=False
    stochastic_tester = tester.StochasticTester(database, 1 / size, privacy_unit, synthetic_data)
    n_sim = 10000
    #test_count(database, stochastic_tester, n_sim, name="figures/count", row_privacy=row_privacy)
    test_sum(database, stochastic_tester, n_sim, distributions=distributions, name="figures/sum", row_privacy=row_privacy)
    # test_avg(database, stochastic_tester, n_sim, distributions=distributions, name="figures/avg", row_privacy=row_privacy)
    # test_joins(database, stochastic_tester, n_sim, name="figures/query_with_join")










MAGENTA_COLOR = '\033[35m'
BLUE_COLOR = '\033[34m'
RESET_COLOR = '\033[0m'


def print_query(query: str) -> None:
    """
    Print a formatted and syntax-highlighted SQL query.

    This function takes an SQL query as input and prints it to the console with syntax
    highlighting applied to certain keywords for better readability.

    Parameters
    ----------
    query: str
        The SQL query to be printed and highlighted.

    Returns
    -------
    None
        This function does not return any value. It directly prints the formatted
        and highlighted SQL query to the console.
    """
    keywords = ["SELECT", "AS", "GROUP BY", "LIMIT", "ORDER BY", "WHERE"]
    colored_query = query
    colored_query = colored_query.replace("WITH", "WITH\n ")
    colored_query = colored_query.replace(" SELECT", "\nSELECT")
    colored_query = colored_query.replace("),", "),\n ")
    for word in keywords:
        colored_query = colored_query.replace(word, MAGENTA_COLOR + word + RESET_COLOR)
    colored_query = colored_query.replace("WITH", BLUE_COLOR + "WITH" + RESET_COLOR)
    print(colored_query)



import typing as t
from .sqlite import SQLite
import pandas as pd

def from_csv(table_name: str, csv_file: str, db_name:str ="my_sqlite.db", ranges: bool=False) -> 'SQLite':
    """
    Construct a PyQrlew Database from a csv file.
    Its uses SQLite for storing the data. The data is cloned into Sqlite.

    Parameters
    ----------
    table_name : str
        Name of the table that is created
    csv_file : str
        Any valid string path is acceptable. The string could be a URL.
    db_name : str, default "my_sqlite.db"
        Name of the file where the database is stored
    ranges: bool
        Query the database for fetching min and max of each column of type integer, float, date/time and string and use then as bounds

    Returns
    -------
    None

    Example
    --------
    >>> import pyqrlew as pq
    >>> qdb = pq.from_csv(
            table_name="heart_data",
            csv_file="https://raw.githubusercontent.com/vserret/datasets/main/heart_data.csv"
        )
    >>> qdb.print_infos_metadata()
        Table: heart_data
         - id, Type: BIGINT
         - age, Type: BIGINT
         - gender, Type: BIGINT
         - height, Type: BIGINT
         - weight, Type: FLOAT
         - ap_hi, Type: BIGINT
         - ap_lo, Type: BIGINT
         - cholesterol, Type: BIGINT
         - gluc, Type: BIGINT
         - smoke, Type: BIGINT
         - alco, Type: BIGINT
         - active, Type: BIGINT
         - cardio, Type: BIGINT
    """
    db = SQLite(db_name, ranges)
    db.load_csv(table_name, csv_file)
    return db

def from_csv_dict(csv_dict: t.Dict[str, str], db_name:str ="my_sqlite.db", ranges: bool=False) -> 'SQLite':
    """
    Construct a PyQrlew Database from a dictionary whose keys are the table names and values the
    corresponding data stored in csv files.
    Its uses SQLite for storing the data. The data is cloned into Sqlite.

    Parameters
    ----------
    data_dict : Dict[str, str]
        For each item, the key is the name of the table that is created
        and the value is the name (path or url) of the csv file where the data is stored
    db_name : str, default "my_sqlite.db"
        Name of the file where the database is stored
    ranges: bool
        Query the database for fetching min and max of each column of type integer, float, date/time and string and use then as bounds

    Returns
    -------
    None

    Example
    --------
    >>> import pyqrlew as pq
    >>> import pandas as pd
    >>> df = pd.DataFrame([[1, "a", True], [2, "b", False]], columns=["col1", "col2", "col3"])
    >>> df.to_csv("table2.csv")
    >>> qdb = pq.from_csv_dict(
        {
            "heart_data": "https://raw.githubusercontent.com/vserret/datasets/main/heart_data.csv",
            "table2": "table2.csv"
        }
    )
    >>> qdb.print_infos_metadata()
        Table: heart_data
        - id, Type: BIGINT
        - age, Type: BIGINT
        - gender, Type: BIGINT
        - height, Type: BIGINT
        - weight, Type: FLOAT
        - ap_hi, Type: BIGINT
        - ap_lo, Type: BIGINT
        - cholesterol, Type: BIGINT
        - gluc, Type: BIGINT
        - smoke, Type: BIGINT
        - alco, Type: BIGINT
        - active, Type: BIGINT
        - cardio, Type: BIGINT

        Table: table2
        - Unnamed: 0, Type: BIGINT
        - col1, Type: BIGINT
        - col2, Type: TEXT
        - col3, Type: BOOLEAN
    """
    db = SQLite(db_name, ranges)
    _ = {
        db.load_csv(table_name, csv_file)
        for table_name, csv_file in csv_dict.items()
    }
    return db

def from_pandas(table_name: str, data: pd.DataFrame, db_name:str ="my_sqlite.db", ranges: bool=False) -> 'SQLite':
    """
    Construct a PyQrlew Database from a pandas DataFrame.
    Its uses SQLite for storing the data. The data is cloned into Sqlite.

    Parameters
    ----------
    table_name : str
        Name of the table that is created
    data : :class:`pandas.DataFrame`
        Data represented as a pandas DataFrame
    db_name : str, default "my_sqlite.db"
        Name of the file where the database is stored
    ranges: bool
        Query the database for fetching min and max of each column of type integer, float, date/time and string and use then as bounds

    Returns
    -------
    None

    Example
    --------
    >>> import pyqrlew as pq
    >>> import pandas as pd
    >>> df = pd.DataFrame([[1, "a", True], [2, "b", False]], columns=["col1", "col2", "col3"])
    >>> qdb = pq.from_pandas("my_table", df)
    >>> qdb.print_infos_metadata()
        Table: my_table
         - col1, Type: BIGINT
         - col2, Type: TEXT
         - col3, Type: BOOLEAN

    """
    db = SQLite(db_name, ranges)
    db.load_pandas(table_name, data)
    return db

def from_pandas_dict(data_dict: t.Dict[str, pd.DataFrame], db_name:str ="my_sqlite.db", ranges: bool=False) -> 'SQLite':
    """
    Construct a PyQrlew Database from a dictionary whose keys are the table names and values the
    corresponding data stored in a pandas DataFrame.
    Its uses SQLite for storing the data. The data is cloned into Sqlite.

    Parameters
    ----------
    data_dict : Dict[str, :class:`pandas.DataFrame`]
        For each item, the key is the name of the table that is created
        and the value is the data represented as a pandas DataFrame
    db_name : str, default "my_sqlite.db"
        Name of the file where the database is stored
    ranges: bool
        Query the database for fetching min and max of each column of type integer, float, date/time and string and use then as bounds

    Returns
    -------
    None

    Example
    --------
    >>> import pyqrlew as pq
    >>> import pandas as pd
    >>> df1 = pd.DataFrame([[1, "a", True], [2, "b", False]], columns=["col1", "col2", "col3"])
    >>> df2 = pd.DataFrame([[1, 2], [3, 4], [5, 6]], columns=["colA", "colB"])
    >>> qdb = pq.from_pandas_dict({"table1": df1, "table2": df2})
    >>> qdb.print_infos_metadata()
        Table: table1
        - col1, Type: BIGINT
        - col2, Type: TEXT
        - col3, Type: BOOLEAN

        Table: table2
        - colA, Type: BIGINT
        - colB, Type: BIGINT
    """
    db = SQLite(db_name, ranges)
    _ = {
        db.load_pandas(table_name, df)
        for table_name, df in data_dict.items()
    }
    return db



from dataclasses import dataclass
import logging
from uuid import uuid4 as generate_uuid
from typing import Optional, Tuple, List, Dict, Union
import json
from sqlalchemy import MetaData, Table, Column, types, select, func, literal, String, ARRAY, case, text
from sqlalchemy.engine import Engine
import pyqrlew as qrl

def dataset_from_database(
    name: str,
    engine: Engine,
    schema_name: Optional[str]=None,
    ranges: bool=False,
    possible_values_threshold: Optional[int]=None
) -> qrl.Dataset:
    """Builds a `Dataset` from a sqlalchemy `Engine`

    Args:
        name (str): Name of the Dataset
        engine (Engine): The sqlalchemy `Engine` to use
        schema_name (Optional[str], optional): The DB schema to use. Defaults to None.
        ranges (bool, optional): Use the actual min and max of the data as ranges. **This is unsafe from a privacy perspective**. Defaults to False.
        possible_values_threshold (Optional[int], optional): Use the actual observed values as range. **This is unsafe from a privacy perspective**. Defaults to None.

    Returns:
        qrl.Dataset: A Qrlew `Dataset` object
    """
    metadata = MetaData()
    metadata.reflect(engine, schema=schema_name)

    def dataset_schema_size() -> Tuple[dict, dict, Optional[dict]]:
        """Return a (dataset, schema) pair or (dataset, schema, size) triplet """
        ds = dataset()
        return (
            ds,
            schema(ds),
            size(ds),
        )

    def dataset() -> dict:
        """Return a (dataset, schema) pair or (dataset, schema, size) triplet """
        return {
            '@type': 'sarus_data_spec/sarus_data_spec.Dataset',
            'uuid': generate_uuid().hex,
            'name': 'Transformed',
            'spec': {
                'transformed': {
                    'transform': generate_uuid().hex,
                    'arguments': [],
                    'named_arguments': {},
                },
            },
            'properties': {},
            'doc': 'This ia a demo dataset for testing purpose',
        }

    def schema(dataset: dict) -> dict:
        tables = {"fields": [table(metadata.tables[name]) for name in metadata.tables]}

        if schema_name is not None:
            tables = {
                "fields": [{
                    'name': schema_name,
                    'type': {
                        'name': 'Union',
                        'union': tables,
                        'properties': {
                            'public_fields': '[]'
                        },
                    }
                }],
            }

        return {
            '@type': 'sarus_data_spec/sarus_data_spec.Schema',
            'uuid': generate_uuid().hex,
            'dataset': dataset['uuid'],
            'name': name,
            'type': {
                # Slugname
                'name': name.lower(),
                'struct': {
                    'fields': [
                        {
                            'name': 'sarus_data',
                            'type': {
                                'name': 'Union',
                                'union': tables,
                                'properties': {
                                    'public_fields': '[]',
                                },
                            },
                        },
                        {
                            'name': 'sarus_weights',
                            'type': {
                                'name': 'Integer',
                                'integer': {
                                    'min': '-9223372036854775808',
                                    'max': '9223372036854775807',
                                    'base': 'INT64',
                                    'possible_values': []
                                },
                                'properties': {}
                            },
                        },
                        {
                            'name': 'sarus_is_public',
                            'type': {
                                'name': 'Boolean',
                                'boolean': {},
                                'properties': {},
                            },
                        },
                        {
                            'name': 'sarus_protected_entity',
                            'type': {
                                'name': 'Id',
                                'id': {
                                    'base': 'STRING',
                                    'unique': False,
                                },
                                'properties': {},
                            },
                        },
                    ],
                },
                'properties': {}
            },
            'protected': {
                'label': 'data',
                'paths': [],
                'properties': {},
            },
            'properties': {
                'max_max_multiplicity': '1',
                'foreign_keys': '',
                'primary_keys': '',
            }
        }

    def table(tab: Table) -> dict:
        min_max_possible_values = compute_min_max_possible_values(tab)
        return {
            'name': tab.name,
            'type': {
                'name': 'Struct',
                'struct': {
                    'fields': [column(col, **min_max_possible_values[col.name]) for col in tab.columns],
                },
                'properties': {},
            }
        }

    def compute_min_max_possible_values(tab: Table) -> Dict[str, Dict[str, Union[str, List[str]]]]:
        """Send 3 SQL queries for loading the bounds"""
        values = {col.name: {'min': None, 'max': None, 'possible_values': []} for col in tab.columns}
        intervals_types = [
            types.Integer, types.BigInteger,
            types.Float, types.Numeric,
            types.String, types.Text, types.Unicode, types.UnicodeText,
            types.Date, types.DateTime, types.Time
        ]
        interval_cols = [
            col for col in tab.columns
            if any(isinstance(col.type, t) for t in intervals_types)
        ]

        if ranges and len(interval_cols) != 0:
            min_query = select(*[func.cast(func.min(col), String).label(col.name) for col in interval_cols]).select_from(tab)
            max_query = select(*[func.cast(func.max(col), String).label(col.name) for col in interval_cols]).select_from(tab)

            with engine.connect() as conn:
                min_results = conn.execute(min_query).fetchone()
                max_results = conn.execute(max_query).fetchone()

            for col, min_val, max_val in zip(interval_cols, min_results, max_results):
                values[col.name]['min'] = min_val
                values[col.name]['max'] = max_val

        if possible_values_threshold is not None and len(interval_cols) != 0:
            tablename = f"\"{tab.name}\"" if tab.schema is None else f"\"{tab.schema}\".\"{tab.name}\""
            values_query = text(
                "SELECT " + ','.join([
                    f"CASE WHEN COUNT(DISTINCT \"{col.name}\") <= {possible_values_threshold} "
                    f"THEN array_agg(DISTINCT CAST(\"{col.name}\" AS Text)) ELSE ARRAY[]::VARCHAR[] "
                    f"END AS \"{col.name}\" "
                    for col in interval_cols
                ])
                + f"FROM {tablename}"
            ) # case very complicate to use with loop

            with engine.connect() as conn:
                values_results = conn.execute(values_query).fetchone()

            for col, possible_values in zip(interval_cols, values_results):
                values[col.name]['possible_values'] = [str(v) for v in possible_values]

        return values

    def column(col: Column, min:Optional[str]=None, max:Optional[str]=None, possible_values:List[str]=[]) -> dict:
        if isinstance(col.type, types.Integer) or isinstance(col.type, types.BigInteger):
            min = '-9223372036854775808' if min is None else min
            max = '9223372036854775807' if max is None else max
            return {
                'name': col.name,
                'type': {
                    'name': 'Integer',
                    'integer': {
                        'base': 'INT64',
                        'min': min,
                        'max': max,
                        'possible_values': possible_values,
                    },
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.Float) or isinstance(col.type, types.Numeric):
            min = '-1.7976931348623157e+308' if min is None else min
            max = '1.7976931348623157e+308' if max is None else max
            return {
                'name': col.name,
                'type': {
                    'name': 'Float64',
                    'float': {
                        'base': 'FLOAT64',
                        'min': min,
                        'max': max,
                        'possible_values': [],
                    },
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.String) or isinstance(col.type, types.Text) or isinstance(col.type, types.Unicode) or isinstance(col.type, types.UnicodeText):

            return {
                'name': col.name,
                'type': {
                    'name': 'Text UTF-8',
                    'text': {
                        'encoding': 'UTF-8',
                        'min': min,
                        'max': max,
                        'possible_values': possible_values,
                    },
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.Boolean):
            return {
                'name': col.name,
                'type': {
                    'name': 'Boolean',
                    'boolean': {},
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.Date) or isinstance(col.type, types.DateTime) or isinstance(col.type, types.Time):
            min = '01-01-01 00:00:00' if min is None else min
            max = '9999-12-31 00:00:00' if max is None else max
            return {
                'name': col.name,
                'type': {
                    'name': 'Datetime',
                    'datetime': {
                        'format': '%Y-%m-%d %H:%M:%S',
                        'min': min,
                        'max': max,
                    },
                    'properties': {},
                },
            }
        else:
            return {
                'name': col.name,
                'type': {
                    'name': 'Type',
                    'type': {},
                    'properties': {},
                },
            }

    def size(dataset: dict) -> dict:
        tables = {'fields': [table_size(metadata.tables[name]) for name in metadata.tables]}
        if schema_name is not None:
            tables = {
                'fields': [
                    {
                        'name': schema_name,
                        'statistics': {
                            'name': 'Union',
                            'union': tables,
                            'properties': {},
                        },
                        'properties': {},
                    },
                ],
            }

        return {
            '@type': 'sarus_data_spec/sarus_data_spec.Size',
            'uuid': generate_uuid().hex,
            'dataset': dataset['uuid'],
            'name': f'{name}_sizes',
            'statistics': {
                'name': 'Union',
                'union': tables,
                'properties': {},
            },
            'properties': {},
        }

    def table_size(tab: Table) -> dict:
        with engine.connect() as conn:
            result = conn.execute(select(func.count()).select_from(tab))
            size = result.scalar()
        multiplicity = 1.0
        return {
            'name': tab.name,
            'statistics': {
                'name': 'Struct',
                'struct': {
                    'fields': [column_size(col, size, multiplicity) for col in tab.columns],
                    'size': str(size),
                    'multiplicity': multiplicity,
                },
                'properties': {},
            }
        }

    def column_size(col: Column, size: int, multiplicity: float) -> dict:
        if isinstance(col.type, types.Integer) or isinstance(col.type, types.BigInteger):
            return {
                'name': col.name,
                'statistics': {
                    "integer": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Integer",
                    "properties": {}
                }
            }
        elif isinstance(col.type, types.Float) or isinstance(col.type, types.Numeric):
            return {
                'name': col.name,
                'statistics': {
                    "float": {
                        "distribution": {
                            "double": {
                                "max": '1.7976931348623157e+308',
                                "min": '-1.7976931348623157e+308',
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Float",
                    "properties": {}
                }
            }
        elif isinstance(col.type, types.String) or isinstance(col.type, types.Text) or isinstance(col.type, types.Unicode) or isinstance(col.type, types.UnicodeText):
            return {
                'name': col.name,
                'statistics': {
                    "text": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Text",
                    "properties": {}
                }
            }

        elif isinstance(col.type, types.Boolean):
            return {
                'name': col.name,
                'statistics': {
                    "boolean": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Boolean",
                    "properties": {}
                }
            }
        elif isinstance(col.type, types.Date) or isinstance(col.type, types.DateTime) or isinstance(col.type, types.Time):
            return {
                'name': col.name,
                'statistics': {
                    "datetime": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Datetime",
                    "properties": {}
                }
            }

        else: # TODO: support more types
            raise NotImplementedError(f"SQL -> Sarus Conversion not supported for {col.type} SQL type")

    # Gather protobufs
    dataset, schema, size = dataset_schema_size()
    # Display when debugging
    logging.debug(json.dumps(dataset))
    logging.debug(json.dumps(schema))
    logging.debug(json.dumps(size))
    # Return the result
    return qrl.Dataset(json.dumps(dataset), json.dumps(schema), json.dumps(size))

# Make it a builder
qrl.Dataset.from_database = dataset_from_database

# Add a useful const
qrl.Dataset.CONSTRAINT_UNIQUE: str = '_UNIQUE_'

# A method to get select a schema
def schema(dataset: qrl.Dataset, schema: str) -> 'Schema':
    return Schema(dataset, schema)

# Add the method
qrl.Dataset.__getattr__ = schema

@dataclass
class Schema:
    dataset: qrl.Dataset
    schema: str

    def __getattr__(self, table: str) -> 'Table':
        return Table(self.dataset, self.schema, table)

@dataclass
class Table:
    dataset: qrl.Dataset
    schema: str
    table: str

    def __getattr__(self, column: str) -> 'Column':
        return Column(self.dataset, self.schema, self.table, column)

    def relation(self) -> qrl.Relation:
        return next(rel for path, rel in self.dataset.relations() if path[1]==self.schema and path[2]==self.table)

@dataclass
class Column:
    dataset: qrl.Dataset
    schema: str
    table: str
    column: str

    def with_range(self, min:float, max:float) -> qrl.Dataset:
        return self.dataset.with_range(self.schema, self.table, self.column, min, max)
    
    def with_possible_values(self, with_possible_values: List[str]) -> qrl.Dataset:
        return self.dataset.with_possible_values(self.schema, self.table, self.column, with_possible_values)
    
    def with_constraint(self, constraint: str) -> qrl.Dataset:
        return self.dataset.with_constraint(self.schema, self.table, self.column, constraint)

    def with_unique_constraint(self) -> qrl.Dataset:
        return self.with_constraint(qrl.Dataset.CONSTRAINT_UNIQUE)
    
    def with_no_constraint(self) -> qrl.Dataset:
        return self.with_constraint(None)


from typing import List

class StochasticTester:
    def __init__(self, database, delta, protected_entity, synthetic_data):
        self.database = database
        self.delta = delta
        self.protected_entity = protected_entity
        self.synthetic_data = synthetic_data

    def compute_dp_relation(self, query, epsilon):
        budget = {"epsilon": epsilon, "delta": self.delta}
        relation = self.database.dataset().relation(query)

        relation_with_privatequeries = relation.rewrite_with_differential_privacy(
            self.database.dataset(),
            self.protected_entity,
            budget,
            self.synthetic_data,
        )
        return relation_with_privatequeries.relation()

    def utility_per_epsilon(self, query: str, epsilons: List[float], n_sim: int):
        print('computing utility per epsilon ...')
        relations = {
            epsilon: self.compute_dp_relation(query, epsilon)
            for epsilon in epsilons
        }
        print('done')
        return {
            epsilon: [
                self.database.eval(dp_relation)[0][0]
                for _ in range(n_sim)
            ]
            for epsilon, dp_relation in relations.items()
        }

    def compute_results(self, query: str, epsilon: float, n_sim: int):
        dp_relation = self.compute_dp_relation(query, epsilon)
        return [
            self.database.eval(dp_relation)[0][0]
            for _ in range(n_sim)
        ]


import os
import numpy as np
import pandas as pd
from pyqrlew.io import PostgreSQL
from pyqrlew import Dialect

# pytest -s tests/test_dataset.py::test_ranges
def test_ranges():
    """Test the consistency of results for queries stored in a file."""
    database = PostgreSQL()
    dataset = database.extract() # load the db        
    new_dataset = dataset.extract.census.age.with_range(20, 42)
    relation = new_dataset.relation('SELECT age FROM extract.census')
    print(relation.schema())
    assert(relation.schema()=='{age: int[20 42]}')

# pytest -s tests/test_dataset.py::test_possible_values
def test_possible_values():
    """Test the consistency of results for queries stored in a file."""
    database = PostgreSQL()
    dataset = database.extract() # load the db        
    new_dataset = dataset.extract.census.workclass.with_possible_values(['Local-gov', 'Private'])
    relation = new_dataset.relation('SELECT workclass FROM extract.census')
    print(relation.schema())
    assert(relation.schema()=='{workclass: str{Local-gov, Private}}')

# pytest -s tests/test_dataset.py::test_constraint
def test_constraint():
    """Test the consistency of results for queries stored in a file."""
    database = PostgreSQL()
    dataset = database.extract() # load the db        
    new_dataset = dataset.extract.census.age.with_unique_constraint()
    relation = new_dataset.relation('SELECT age FROM extract.census')
    print(relation.schema())
    assert(relation.schema()=='{age: int[20 90] (UNIQUE)}')

# pytest -s tests/test_dataset.py::test_relation
def test_relation():
    """Test the consistency of results for queries stored in a file."""
    database = PostgreSQL()
    dataset = database.extract() # load the db        
    relation = dataset.extract.census.relation()
    print(relation.schema())



import os
import sqlalchemy
from qrlew_datasets.database import Database
import pyqrlew as qrl
from pyqrlew.io.dataset import dataset_from_database
import pandas as pd

dtype_to_sqlalchemy = {
    'object': sqlalchemy.String,
    'int64': sqlalchemy.Integer,
    'float64': sqlalchemy.Float,
    'bool': sqlalchemy.Boolean,
    'datetime64[ns]': sqlalchemy.DateTime,
    'timedelta64[ns]': sqlalchemy.Interval,
    'category': sqlalchemy.String
}


class SQLite(Database):
    """
    A class representing an SQLite database with PyQrlew integration.

    This class provides methods for interacting with an SQLite database using the
    PyQrlew library.

    Parameters
    ----------
        db_file: str
            The path to the SQLite database file.

    Attributes
    ----------
        db_file: str
            The path to the SQLite database file.

    Example
    ----------
        >>> import pandas as pd
        >>> qdb = SQLite("my_database.db")
        >>> df = pd.DataFrame([[1, "a", True], [2, "b", False]], columns=["col1", "col2", "col3"])
        >>> qdb.load_pandas("table1", df)
        >>> qdb.print_infos_metadata()
        Table: table1
        - col1, Type: BIGINT
        - col2, Type: TEXT
        - col3, Type: BOOLEAN
        >>> qdb.dataset().relations()
        [(['table1'], <Relation at 0x120f47120>)]
        >>> query = "SELECT SUM(col1) FROM table1"
        >>> qdb.execute(query)
        [(3,)]
        >>> relation = qdb.dataset().sql(query)
        >>> qdb.eval(relation)
        [(3,)]
    """

    def __init__(self, db_file:str, ranges: bool) -> None:
        """
        Initialize an SQLite instance.

        Parameters
        ----------
            db_file: str
                The path to the SQLite database file.
            ranges: bool
                Query the database for fetching min and max of each column of type integer, float, date/time and string and use then as bounds
        Returns
        ----------
            None
        """
        self.db_file = db_file
        self.ranges = ranges
        self.possible_values_threshold = None # Array no implemented in SQLite

    def load_csv(self, table_name: str, csv_file: str) -> None:
        """
        Create a table in the SQLite database from data stored in a csv file.

        Parameters
        ----------
            table_name: str
                Name of the created table

            csv_file: str
                Path or URL to the CSV file containing the data.

        Returns
        ----------
            None

        """
        df = pd.read_csv(csv_file)
        self.load_pandas(table_name, df)

    def load_pandas(self, table_name: str, df: pd.DataFrame) -> None:
        """
        Create a table in the SQLite database from data stored in a Pandas DataFrame.

        Parameters
        ----------
            table_name: str
                Name of the created table

            df: :class:`pandas.DataFrame`
                Data represented as a pandas DataFrame

        Returns
        ----------
            None

        """
        column_types = {
            column: dtype_to_sqlalchemy[str(df[column].dtype)]
            for column in df.columns
        }
        metadata = sqlalchemy.MetaData()
        columns = [
            sqlalchemy.Column(col_name, col_type)
            for col_name, col_type in column_types.items()
        ]
        _ = sqlalchemy.Table(table_name, metadata, *columns)
        engine = self.engine()
        metadata.create_all(engine)
        df.to_sql(table_name, engine, if_exists='replace', index=False)

    def url(self) -> str:
        """Get the URL string of the SQLite database."""
        return f'sqlite:///{self.db_file}'

    def remove(self) -> None:
        """Remove the SQLite database file from the filesystem."""
        os.remove(self.db_file)

    def engine(self) -> sqlalchemy.engine.Engine:
        """Create and return an SQLAlchemy Engine instance for the SQLite database."""
        return sqlalchemy.create_engine(self.url(), echo=True)

    def eval(self, relation: qrl.Relation) -> list:
        """Convert a PyQrlew relation into a sql query string, send it to the database and return the result."""
        return self.execute(relation.to_query())

    def execute(self, query: str) -> list:
        """Execute a raw SQL query and return the result."""
        with self.engine().connect() as conn:
            result = conn.execute(sqlalchemy.text(query)).all()
        return result

    def dataset(self) -> qrl.Dataset:
        """Create and return a PyQrlew Dataset linked to the SQLite database."""
        return dataset_from_database(self.db_file, self.engine(), ranges=self.ranges, possible_values_threshold=self.possible_values_threshold)

    def print_infos_metadata(self) -> None:
        """Print metadata information about tables and columns in the database."""
        metadata = sqlalchemy.MetaData()
        metadata.reflect(self.engine(), schema=None)
        for table_name, table in metadata.tables.items():
            print(f"\nTable: {table_name}")
            for column in table.c:
                print(f" - {column.name}, Type: {column.type}")



import shutil
from pathlib import Path
from sqlalchemy import select, column, table, text
from qrlew_datasets.files import SQL
from qrlew_datasets.databases import PostgreSQL as EmptyPostgreSQL
from pyqrlew.io.dataset import dataset_from_database
import pyqrlew as qrl

NAME: str = 'pyqrlew-db'
USER: str = 'postgres'
PASSWORD: str = 'pyqrlew-db'
PORT: str = 5433

POSSIBLE_VALUES_THRESHOLD: int = 20

class PostgreSQL(EmptyPostgreSQL):
    """_summary_

    Args:
        EmptyPostgreSQL (_type_): _description_
    """
    def __init__(self, name=NAME, user=USER, password=PASSWORD, port=PORT) -> None:
        """_summary_

        Args:
            name (_type_, optional): _description_. Defaults to NAME.
            user (_type_, optional): _description_. Defaults to USER.
            password (_type_, optional): _description_. Defaults to PASSWORD.
            port (_type_, optional): _description_. Defaults to PORT.
        """
        super().__init__(name, user, password, port)

    def load_resource(self, schema: str, src: Path) -> 'PostgreSQL':
        """_summary_

        Args:
            schema (str): _description_
            src (Path): _description_

        Returns:
            PostgreSQL: _description_
        """
        with self.engine().connect() as conn:
            res = list(conn.execute(select(column('schema_name')).select_from(table('schemata', schema='information_schema')).where(column('schema_name') == schema)))
            schema_exists = len(res)==1
        if not schema_exists:
            dst = Path('/tmp') / schema
            shutil.copyfile(src, dst)
            self.load(dst)
        return self

    def load_extract(self) -> 'PostgreSQL':
        """_summary_

        Returns:
            PostgreSQL: _description_
        """
        return self.load_resource('extract', SQL('extract').local())

    def load_financial(self) -> 'PostgreSQL':
        """_summary_

        Returns:
            PostgreSQL: _description_
        """
        return self.load_resource('financial', SQL('financial').local())

    def load_hepatitis(self) -> 'PostgreSQL':
        """_summary_

        Returns:
            PostgreSQL: _description_
        """
        return self.load_resource('hepatitis_std', SQL('hepatitis').local())

    def load_imdb(self) -> 'PostgreSQL':
        """_summary_

        Returns:
            PostgreSQL: _description_
        """
        return self.load_resource('imdb_ijs', SQL('imdb').local())

    def load_retail(self) -> 'PostgreSQL':
        """_summary_

        Returns:
            PostgreSQL: _description_
        """
        return self.load_resource('retail', SQL('retail').local())

    def extract(self) -> qrl.Dataset:
        """_summary_

        Returns:
            qrl.Dataset: _description_
        """
        self.load_extract()
        return dataset_from_database('extract', self.engine(), 'extract', ranges=True, possible_values_threshold=POSSIBLE_VALUES_THRESHOLD)

    def financial(self) -> qrl.Dataset:
        """_summary_

        Returns:
            qrl.Dataset: _description_
        """
        self.load_financial()
        return dataset_from_database('financial', self.engine(), 'financial', ranges=True, possible_values_threshold=POSSIBLE_VALUES_THRESHOLD)

    def hepatitis(self) -> qrl.Dataset:
        """_summary_

        Returns:
            qrl.Dataset: _description_
        """
        self.load_hepatitis()
        return dataset_from_database('hepatitis_std', self.engine(), 'hepatitis_std', ranges=True, possible_values_threshold=POSSIBLE_VALUES_THRESHOLD)

    def imdb(self) -> qrl.Dataset:
        """_summary_

        Returns:
            qrl.Dataset: _description_
        """
        self.load_imdb()
        return dataset_from_database('imdb_ijs', self.engine(), 'imdb_ijs', ranges=True, possible_values_threshold=POSSIBLE_VALUES_THRESHOLD)

    def retail(self) -> qrl.Dataset:
        """_summary_

        Returns:
            qrl.Dataset: _description_
        """
        self.load_retail()
        return dataset_from_database('retail', self.engine(), 'retail', ranges=True, possible_values_threshold=POSSIBLE_VALUES_THRESHOLD)
